@@ -191,8 +191,9 @@ pub struct Encoder<W: Write> {
 
     sampling_factor: SamplingFactor,
 
-    // A non zero value enables progressive mode
-    progressive_scans: u8,
+    progressive_scans: Option<u8>,
+
+    restart_interval: Option<u16>,
 
     optimize_huffman_table: bool,
 
@@ -229,7 +230,8 @@ impl<W: Write> Encoder<W> {
             quantization_tables,
             huffman_tables,
             sampling_factor,
-            progressive_scans: 0,
+            progressive_scans: None,
+            restart_interval: None,
             optimize_huffman_table: false,
             app_segments: Vec::new(),
         }
@@ -263,9 +265,9 @@ impl<W: Write> Encoder<W> {
     /// Use [set_progressive_scans](Encoder::set_progressive_scans) to use a different number of scans
     pub fn set_progressive(&mut self, progressive: bool) {
         self.progressive_scans = if progressive {
-            4
+            Some(4)
         } else {
-            0
+            None
         };
     }
 
@@ -278,15 +280,28 @@ impl<W: Write> Encoder<W> {
     /// If number of scans is not within valid range
     pub fn set_progressive_scans(&mut self, scans: u8) {
         assert!((2..=64).contains(&scans), "Invalid number of scans: {}", scans);
-        self.progressive_scans = scans;
+        self.progressive_scans = Some(scans);
     }
 
     /// Return number of progressive scans if progressive encoding is enabled
     pub fn progressive_scans(&self) -> Option<u8> {
-        match self.progressive_scans {
-            0 => None,
-            scans => Some(scans)
-        }
+        self.progressive_scans
+    }
+
+    /// Set restart interval
+    ///
+    /// Set numbers of MCUs between restart markers.
+    pub fn set_restart_interval(&mut self, interval: u16) {
+        self.restart_interval = if interval == 0 {
+            None
+        } else {
+            Some(interval)
+        };
+    }
+
+    /// Return the restart interval
+    pub fn restart_interval(&self) -> Option<u16> {
+        self.restart_interval
     }
 
     /// Set if optimized huffman table should be created
@@ -413,8 +428,8 @@ impl<W: Write> Encoder<W> {
             self.writer.write_segment(Marker::APP(*nr), data)?;
         }
 
-        if self.progressive_scans != 0 {
-            self.encode_image_progressive(image)?;
+        if let Some(scans) = self.progressive_scans {
+            self.encode_image_progressive(image, scans)?;
         } else if self.optimize_huffman_table || !self.sampling_factor.supports_interleaved() {
             self.encode_image_sequential(image)?;
         } else {
@@ -466,7 +481,7 @@ impl<W: Write> Encoder<W> {
     }
 
     fn write_frame_header<I: ImageBuffer>(&mut self, image: &I) -> Result<(), EncodingError> {
-        self.writer.write_frame_header(image.width() as u16, image.height() as u16, &self.components, self.progressive_scans != 0)?;
+        self.writer.write_frame_header(image.width() as u16, image.height() as u16, &self.components, self.progressive_scans.is_some())?;
 
         self.writer.write_quantization_segment(0, &self.quantization_tables[0])?;
         self.writer.write_quantization_segment(1, &self.quantization_tables[1])?;
@@ -497,10 +512,13 @@ impl<W: Write> Encoder<W> {
             )?;
         }
 
+        if let Some(restart_interval) = self.restart_interval {
+            self.writer.write_dri(restart_interval)?;
+        }
+
         Ok(())
     }
 
-
     fn init_rows(&mut self, buffer_size: usize) -> [Vec<u8>; 4] {
 
         // To simplify the code and to give the compiler more infos to optimize stuff we always initialize 4 components
@@ -567,6 +585,10 @@ impl<W: Write> Encoder<W> {
 
         let mut prev_dc = [0i16; 4];
 
+        let restart_interval = self.restart_interval.unwrap_or(0);
+        let mut restarts = 0;
+        let mut restarts_to_go = restart_interval;
+
         for block_y in 0..num_rows {
             for r in &mut row {
                 r.clear();
@@ -585,6 +607,16 @@ impl<W: Write> Encoder<W> {
             }
 
             for block_x in 0..num_cols {
+                if restart_interval > 0 && restarts_to_go == 0 {
+                    self.writer.finalize_bit_buffer()?;
+                    self.writer.write_marker(Marker::RST((restarts % 8) as u8))?;
+
+                    prev_dc[0] = 0;
+                    prev_dc[1] = 0;
+                    prev_dc[2] = 0;
+                    prev_dc[3] = 0;
+                }
+
                 for (i, component) in self.components.iter().enumerate() {
                     for v_offset in 0..component.vertical_sampling_factor as usize {
                         for h_offset in 0..component.horizontal_sampling_factor as usize {
@@ -612,6 +644,15 @@ impl<W: Write> Encoder<W> {
                         }
                     }
                 }
+
+                if restart_interval > 0 {
+                    if restarts_to_go == 0 {
+                        restarts_to_go = restart_interval;
+                        restarts += 1;
+                        restarts &= 7;
+                    }
+                    restarts_to_go -= 1;
+                }
             }
         }
 
@@ -634,11 +675,22 @@ impl<W: Write> Encoder<W> {
         self.write_frame_header(&image)?;
 
         for (i, component) in self.components.iter().enumerate() {
+            let restart_interval = self.restart_interval.unwrap_or(0);
+            let mut restarts = 0;
+            let mut restarts_to_go = restart_interval;
+
             self.writer.write_scan_header(&[component], None)?;
 
             let mut prev_dc = 0;
 
             for block in &blocks[i] {
+                if restart_interval > 0 && restarts_to_go == 0 {
+                    self.writer.finalize_bit_buffer()?;
+                    self.writer.write_marker(Marker::RST((restarts % 8) as u8))?;
+
+                    prev_dc = 0;
+                }
+
                 self.writer.write_block(
                     &block,
                     prev_dc,
@@ -647,6 +699,15 @@ impl<W: Write> Encoder<W> {
                 )?;
 
                 prev_dc = block[0];
+
+                if restart_interval > 0 {
+                    if restarts_to_go == 0 {
+                        restarts_to_go = restart_interval;
+                        restarts += 1;
+                        restarts &= 7;
+                    }
+                    restarts_to_go -= 1;
+                }
             }
 
             self.writer.finalize_bit_buffer()?;
@@ -661,6 +722,7 @@ impl<W: Write> Encoder<W> {
     fn encode_image_progressive<I: ImageBuffer>(
         &mut self,
         image: I,
+        scans: u8,
     ) -> Result<(), EncodingError> {
         let blocks = self.encode_blocks(&image);
 
@@ -675,9 +737,20 @@ impl<W: Write> Encoder<W> {
         for (i, component) in self.components.iter().enumerate() {
             self.writer.write_scan_header(&[component], Some((0, 0)))?;
 
+            let restart_interval = self.restart_interval.unwrap_or(0);
+            let mut restarts = 0;
+            let mut restarts_to_go = restart_interval;
+
             let mut prev_dc = 0;
 
             for block in &blocks[i] {
+                if restart_interval > 0 && restarts_to_go == 0 {
+                    self.writer.finalize_bit_buffer()?;
+                    self.writer.write_marker(Marker::RST((restarts % 8) as u8))?;
+
+                    prev_dc = 0;
+                }
+
                 self.writer.write_dc(
                     block[0],
                     prev_dc,
@@ -685,13 +758,22 @@ impl<W: Write> Encoder<W> {
                 )?;
 
                 prev_dc = block[0];
+
+                if restart_interval > 0 {
+                    if restarts_to_go == 0 {
+                        restarts_to_go = restart_interval;
+                        restarts += 1;
+                        restarts &= 7;
+                    }
+                    restarts_to_go -= 1;
+                }
             }
 
             self.writer.finalize_bit_buffer()?;
         }
 
         // Phase 2: AC scans
-        let scans = self.progressive_scans as usize - 1;
+        let scans = scans as usize - 1;
 
         let values_per_scan = 64 / scans;
 
@@ -705,15 +787,33 @@ impl<W: Write> Encoder<W> {
             };
 
             for (i, component) in self.components.iter().enumerate() {
+                let restart_interval = self.restart_interval.unwrap_or(0);
+                let mut restarts = 0;
+                let mut restarts_to_go = restart_interval;
+
                 self.writer.write_scan_header(&[component], Some((start as u8, end as u8 - 1)))?;
 
                 for block in &blocks[i] {
+                    if restart_interval > 0 && restarts_to_go == 0 {
+                        self.writer.finalize_bit_buffer()?;
+                        self.writer.write_marker(Marker::RST((restarts % 8) as u8))?;
+                    }
+
                     self.writer.write_ac_block(
                         &block,
                         start,
                         end,
                         &self.huffman_tables[component.ac_huffman_table as usize].1,
                     )?;
+
+                    if restart_interval > 0 {
+                        if restarts_to_go == 0 {
+                            restarts_to_go = restart_interval;
+                            restarts += 1;
+                            restarts &= 7;
+                        }
+                        restarts_to_go -= 1;
+                    }
                 }
 
                 self.writer.finalize_bit_buffer()?;
@@ -727,8 +827,13 @@ impl<W: Write> Encoder<W> {
         let width = image.width();
         let height = image.height();
 
-        let num_cols = ceil_div(usize::from(width), 8);
-        let num_rows = ceil_div(usize::from(height), 8);
+        let (max_h_sampling, max_v_sampling) = self.get_max_sampling_size();
+
+        let num_cols = ceil_div(usize::from(width), 8 * max_h_sampling) * max_h_sampling;
+        let num_rows = ceil_div(usize::from(height), 8 * max_v_sampling) * max_v_sampling;
+
+        debug_assert!(num_cols > 0);
+        debug_assert!(num_rows > 0);
 
         let buffer_width = (num_cols * 8) as usize;
         let buffer_size = (num_cols * num_rows * 64) as usize;
@@ -748,8 +853,6 @@ impl<W: Write> Encoder<W> {
 
         let mut blocks: [Vec<_>; 4] = self.init_block_buffers(buffer_size / 64);
 
-        let (max_h_sampling, max_v_sampling) = self.get_max_sampling_size();
-
         for (i, component) in self.components.iter().enumerate() {
             let h_scale = max_h_sampling as usize / component.horizontal_sampling_factor as usize;
             let v_scale = max_v_sampling as usize / component.vertical_sampling_factor as usize;
@@ -757,6 +860,8 @@ impl<W: Write> Encoder<W> {
             let cols = num_cols as usize / h_scale;
             let rows = num_rows as usize / v_scale;
 
+            debug_assert!(cols > 0);
+            debug_assert!(rows > 0);
 
             for block_y in 0..rows {
                 for block_x in 0..cols {
@@ -820,10 +925,17 @@ impl<W: Write> Encoder<W> {
             let mut ac_freq = [0u32; 257];
             ac_freq[256] = 1;
 
+            let mut had_ac = false;
+            let mut had_dc = false;
+
             for (i, component) in self.components.iter().enumerate() {
                 if component.dc_huffman_table == table {
+                    had_dc = true;
+
                     let mut prev_dc = 0;
 
+                    debug_assert!(!blocks[i].is_empty());
+
                     for block in &blocks[i] {
                         let value = block[0];
                         let diff = value - prev_dc;
@@ -836,8 +948,10 @@ impl<W: Write> Encoder<W> {
                 }
 
                 if component.ac_huffman_table == table {
-                    if self.progressive_scans > 0 {
-                        let scans = self.progressive_scans as usize - 1;
+                    had_ac = true;
+
+                    if let Some(scans) = self.progressive_scans {
+                        let scans = scans as usize - 1;
 
                         let values_per_scan = 64 / scans;
 
@@ -850,6 +964,8 @@ impl<W: Write> Encoder<W> {
                                 (scan + 1) * values_per_scan
                             };
 
+                            debug_assert!(!blocks[i].is_empty());
+
                             for block in &blocks[i] {
                                 let mut zero_run = 0;
 
@@ -904,6 +1020,9 @@ impl<W: Write> Encoder<W> {
                 }
             }
 
+            assert!(had_dc, "Missing DC data for table {}", table);
+            assert!(had_ac, "Missing AC data for table {}", table);
+
             self.huffman_tables[table as usize] = (
                 HuffmanTable::new_optimized(dc_freq),
                 HuffmanTable::new_optimized(ac_freq)
